from itertools import imap,count,chain
from weakref import ref

class Table(list):
    """
    
    """
    def __init__(self, headers=None, types=None, colcount=None):
        '''
        
        @param headers: the list of column header.
        
                        if this parametter is C{None}, C{colcount}
                        parametter must be set.
                        
        @type headers: C{list}, C{tuple} or and iterable object
        
        @param types: the list of data type associated to each column.
        
                      If this parametter is specified its length must be
                      equal to the C{headers} length or to C{colcount}.
                      
        @type types: C{list}, C{tuple} or and iterable object
        
        @param colcount: number of column in the created table.
                         
                         If C{headers} parametter is not C{None} this
                         parametter is ignored
                          
        @type colcount: int
        '''
        
        assert headers is not None or colcount is not None,\
            'headers or colcount parametter must be not None value'
            
        if headers is None:
            headers = tuple('Col_%d' % x for x in xrange(colcount))
        
        self.headers = headers
        self.types   = types

        self.likedTo=set()
        
                
    def _castValue(t,x):
        '''
        Cast a value to a specified type, with exception of
        C{None} values that are returned without cast.
        
        @param t: the destination type
        @type t: C{type}
        @param x: the value to cast
        
        @return: the casted value or C{None}
        
        '''
        if x is None or t is None:
            return x
        else:
            return t(x)
        
    _castValue=staticmethod(_castValue)

    def _castRow(self,value):
        
        if not isinstance(value, list,dict):
            value=[x for x in value]
        
        if isinstance(value,list):
            assert len(value)==len(self.header), \
                   'values has not good length'
            if self.type is not None:
                value=[Table._castValue(t, x)
                       for t,x in imap(None,self.type,value)]
                
        elif isinstance(value,dict):
            lvalue = [None] * len(self.header)
            
            for k,v in value.items():
                try:
                    if self.type is not None:
                        lvalue[self._hindex[k]]=Table._castValue(self.types[self._hindex[k]], v)
                    else:
                        lvalue[self._hindex[k]]=v
                except KeyError:
                    info('%s is not a table column' % k)
                    
            value=lvalue
        else:
            raise TypeError
        
        return value
        

    def __setitem__ (self,key,value):
        '''
        
        @param key:
        @type key: C{int}, C{slice} or C{str}
        @param value:
        @type value:
        '''
        if self.isRow:            
            list.__getitem__(self,0)[key]=value
            
        if isintance(key,int):
            value = self._castRow(value)
            list.__setitem__(self,key,value)
        elif isinstance(key,slice):
            value = [self._castRow(x) for x in value]
            list.__setitem__(self,key,value)
        elif isinstance(key,str):
            i = self._hindex[key]
            if self.types is not None:
                t = self.types[i]
            else:
                t=None
            value = [Table._castValue(t, x)
                     for x in value]
            assert len(value)==len(self),\
                "value has not the good length"
            for row,v in imap(None,self,value):
                row[i]=v                    
        else:
            raise TypeError
        
    def __getitem__(self,key):
        '''
        this function has different comportements depending
        of the data type of C{key} and the table used.
        
        @param key: description of the table part to return
        @type key: C{int}, C{slice} or C{str}
        
        @return: return a subpart of the table.
        '''
        if self.isRow:
            if isinstance(key,(int,slice)):
                return list.__getitem__(self,0)[key]
            else:
                raise TypeError              
        elif self.isColumn:
            if isinstance(key,int):
                return list.__getitem__(self,key)[0]
            elif isinstance(key,slice):
                return [x[0] for x in list.__getitem__(self,key)]
            else:
                raise TypeError
        else:
            if isinstance(key,int):
                if key >= len(self):
                    raise IndexError
                key=slice(key,key+1)
            
            if isinstance(key,slice):
                newtable=Table(self.headers,self.types)
                rep = list.__getitem__(self,key)
                list.__setitem__(newtable,slice(None),rep)
                newtable.likedTo.add(self)
                
            elif isinstance(key,str):
                i = self._hindex[key]
                colname = (key,)
                if self.types is not None:
                    coltype = (self.types[i],)
                else:
                    coltype = None
                newtable=Table(colname,self.types)
                list.__setitem__(newtable,
                                 slice(None),
                                 ([x[i]] for x in self))
            else:
                raise TypeError
                                                         
            return newtable
        
    def __getslice__(self,x,y):
        return self.__getitem__(slice(x,y))
    
    def __iter__(self):
        return TableIterator(self)
    
    def __hash__(self):
        return id(self)
    
    def __add__(self,itable):
        return concatTables(self,itable)
    
    def _setTypes(self,types):
        if types is not None and not isinstance(type,tuple):
            types = tuple(x for x in types)
            
        assert types is None or len(types)==len(self._headers)
        
        self._types = types
        
        if types is not None:
            for row in self:
                for i in xrange(len(row)):
                    row[i]=Table._castValue(types[i], row[i])
                    
    def _getTypes(self):
        return self._types
    
    types = property(_getTypes,_setTypes)
    
    def _getHeaders(self):
        return self._headers
    
    def _setHeaders(self,headers):
        if not isinstance(headers, tuple):
            headers = tuple(x for x in headers)

        self._hindex = dict((k,i) for i,k in imap(None,count(),headers))
        self._headers=headers
        
    headers=property(_getHeaders,_setHeaders)   
    
    def _isColumn(self):
        return len(self._headers)==1
    
    def _isRow(self):
        return len(self)==1
    
    isColumn = property(_isColumn,None)
    '''
    True is the table has only one column
    
    @rtype: bool
    '''
    
    
    isRow = property(_isRow,None) 
    '''
    True if the table has only one row
    
    @rtype: bool
    '''    
            
            
        


class iTableIterator(object):
    
    def _getHeaders(self):
        raise NotImplemented
    
    def _getTypes(self):
        raise NotImplemented
    
    def __iter__(self):
        return self
    
    headers = property(_getHeaders,None,None)
    types   = property(_getTypes,None,None)
    
    def columnIndex(self,name):
        if isinstance(name,str):
            return self._reference.headers.index(name)
        elif isinstance(name,int):
            lh = len(self._reference.headers)
            if name < lh and name >=0:
                return name
            elif name < 0 and name >= -lh:
                return lh - name
            raise IndexError
        raise TypeError
    
    def next(self):
        raise NotImplemented
    

class TableIterator(iTableIterator):
    
    def __init__(self,table):
        if not isinstance(table,Table):
            raise TypeError
        
        self._reftable=table
        self._i=0
        
    def _getHeaders(self):
        return self._reftable.headers
    
    def _getTypes(self):
        return self._reftable.types
    
    def columnIndex(self,name):
        if isinstance(name,str):
            return self._reftable._hindex[name]
        elif isinstance(name,int):
            lh = len(self._reftable._headers)
            if name < lh and name >=0:
                return name
            elif name < 0 and name >= -lh:
                return lh - name
            raise IndexError
        raise TypeError
    
    
    def rewind(self):
        i=0
    
    def next(self):
        if self._i < len(self._reftable):
            rep=list.__getitem__(self._reftable,self._i)
            self._i+=1
            return rep
        else:
            raise StopIteration
        
    headers = property(_getHeaders,None,None)
    types   = property(_getTypes,None,None)


class ProjectionIterator(iTableIterator):
    
    def __init__(self,tableiterator,*cols):
        self._reference = iter(tableiterator)
        
        assert isinstance(self._reference, iTableIterator)

        self._selected = tuple(self._reference.columnIndex(x)
                          for x in cols)
        self._headers = tuple(self._reference.headers[x] 
                         for x in self._selected)
        
        if self._reference.types is not None:
            self._types= tuple(self._reference.types[x] 
                         for x in self._selected)
        else:
            self._types=None
                

    def _getHeaders(self):
        return self._headers
    
    def _getTypes(self):
        return self._types
    
    headers = property(_getHeaders,None,None)
    types   = property(_getTypes,None,None)
    
    def next(self):
        value = self._reference.next()
        value = [value[x] for x in self._selected]
        return value
    
class SelectionIterator(iTableIterator):
    def __init__(self,tableiterator,**conditions):
        self._reference = iter(tableiterator)
        
        assert isinstance(self._reference, iTableIterator)
        
        self._conditions=dict((self._reference.columnIndex(i),c) 
                              for i,c in conditions.iteritems())
    
    def _checkCondition(self,row):
        return reduce(lambda x,y : x and y,
                      (bool(self._conditions[i](row[i]))
                       for i in self._conditions),
                       True)
    
    def _getHeaders(self):
        return self._reference.headers
    
    def _getTypes(self):
        return self._reference.types
        
    def next(self):
        row = self._reference.next()
        while not self._checkCondition(row):
            row = self._reference.next()
        return row
            

    headers = property(_getHeaders,None,None)
    types   = property(_getTypes,None,None)
    
    
class UnionIterator(iTableIterator):
    def __init__(self,*itables):
        self._itables=[iter(x) for x in itables]
        self._types = self._itables[0].types
        self._headers = self._itables[0].headers
        
        assert reduce(lambda x,y: x and y,
                      (    isinstance(z,iTableIterator) 
                       and len(z.headers)==len(self._headers)
                       for z in self._itables),
                    True)
        
        self._iterator = chain(*self._itables)
        
    def _getHeaders(self):
        return self._headers
    
    def _getTypes(self):
        return self._types
    
    def next(self):
        value = self._iterator.next()
        if self._types is not None:
            value = [Table._castValue(t, x)
                     for t,x in imap(None,value,self._types)]
        return value
    
    headers = property(_getHeaders,None,None)
    types   = property(_getTypes,None,None)
    
    
def tableFactory(tableiterator):
    tableiterator = iter(tableiterator)
    assert isinstance(tableiterator, iTableIterator)
    
    newtable = Table(tableiterator.headers,
                     tableiterator.types)
    
    for r in tableiterator:
        list.append(newtable,r)
        
    return newtable

def projectTable(tableiterator,*cols):
    return tableFactory(ProjectionIterator(tableiterator,*cols))

def subTable(tableiterator,**conditions):
    return tableFactory(SelectionIterator(tableiterator,**conditions))
   
def concatTables(*itables):
    '''
    Concatene severals tables.
    
    concatenation is done using the L{UnionIterator<UnionIterator>}
    
    @type itables: iTableIterator or Table
    
    @return: a new Table
    @rtype: c{Table}
    
    @see: L{UnionIterator<UnionIterator>}
    '''
    return tableFactory(UnionIterator(*itables))
   
class TableIteratorAsDict(object):
    
    def __init__(self,tableiterator):
        self._reference = iter(tableiterator)
        
        assert isinstance(self._reference, iTableIterator)
        
        self._headers  = self._reference.headers
        self._types     = self._reference.types
        if self._types is not None:
            self._types = dict((n,t) 
                               for n,t in imap(None,self._headers,self._types))
        
    def __iter__(self):
        return self
    
    def next(self):
        value = self._reference.next()
        return dict((n,t)
                    for n,t in imap(None,self._headers,value))
        
    def _getHeaders(self):
        return self._headers
    
    def _getTypes(self):
        return self._types
    
    headers = property(_getHeaders,None,None)
    types   = property(_getTypes,None,None)
     